@@ -29,7 +29,6 @@
 #include <gio/gio.h>
 #include <gtk/gtk.h>
 #include <webkit/webkit.h>
-#include <webkit/webkitwebresource.h>
 
 #include "yelp-debug.h"
 #include "yelp-docbook-document.h"
@@ -56,9 +55,11 @@ static void        yelp_view_set_property         (GObject            *object,
 
 static void        view_scrolled                  (GtkAdjustment      *adjustment,
                                                    YelpView           *view);
-static void        view_scroll_adjustments        (YelpView           *view,
-                                                   GtkAdjustment      *hadj,
-                                                   GtkAdjustment      *vadj,
+static void        view_set_hadjustment           (YelpView           *view,
+                                                   GParamSpec         *pspec,
+                                                   gpointer            data);
+static void        view_set_vadjustment           (YelpView           *view,
+                                                   GParamSpec         *pspec,
                                                    gpointer            data);
 static void        popup_open_link                (GtkMenuItem        *item,
                                                    YelpView           *view);
@@ -68,9 +69,17 @@ static void        popup_save_image               (GtkMenuItem        *item,
                                                    YelpView           *view);
 static void        popup_send_image               (GtkMenuItem        *item,
                                                    YelpView           *view);
+static void        popup_copy_code                (GtkMenuItem        *item,
+                                                   YelpView           *view);
+static void        popup_save_code                (GtkMenuItem        *item,
+                                                   YelpView           *view);
 static void        view_populate_popup            (YelpView           *view,
                                                    GtkMenu            *menu,
                                                    gpointer            data);
+static void        view_script_alert              (YelpView           *view,
+                                                   WebKitWebFrame     *frame,
+                                                   gchar              *message,
+                                                   gpointer            data);
 static gboolean    view_navigation_requested      (WebKitWebView             *view,
                                                    WebKitWebFrame            *frame,
                                                    WebKitNetworkRequest      *request,
@@ -160,6 +169,20 @@ G_DEFINE_TYPE (YelpView, yelp_view, WEBKIT_TYPE_WEB_VIEW);
 
 static WebKitWebSettings *websettings;
 
+typedef struct _YelpActionEntry YelpActionEntry;
+struct _YelpActionEntry {
+    GtkAction               *action;
+    YelpViewActionValidFunc  func;
+    gpointer                 data;
+};
+action_entry_free (YelpActionEntry *entry)
+{
+    if (entry == NULL)
+        return;
+    g_object_unref (entry->action);
+    g_free (entry);
+}
+
 typedef struct _YelpBackEntry YelpBackEntry;
 struct _YelpBackEntry {
     YelpUri *uri;
@@ -194,7 +217,11 @@ struct _YelpViewPrivate {
     gulong         hadjuster;
 
     gchar         *popup_link_uri;
+    gchar         *popup_link_text;
     gchar         *popup_image_uri;
+    WebKitDOMNode *popup_code_node;
+    WebKitDOMNode *popup_code_title;
+    gchar         *popup_code_text;
 
     YelpViewState  state;
 
@@ -204,13 +231,15 @@ struct _YelpViewPrivate {
     gchar         *page_desc;
     gchar         *page_icon;
 
-    GList *back_list;
-    GList *back_cur;
-    gboolean back_load;
+    GList          *back_list;
+    GList          *back_cur;
+    gboolean        back_load;
 
     GtkActionGroup *action_group;
 
-    gint           navigation_requested;
+    GSList         *link_actions;
+
+    gint            navigation_requested;
 };
 
 #define TARGET_TYPE_URI_LIST     "text/uri-list"
@@ -234,10 +263,14 @@ yelp_view_init (YelpView *view)
                           G_CALLBACK (view_navigation_requested), NULL);
     g_signal_connect (view, "resource-request-starting",
                       G_CALLBACK (view_resource_request), NULL);
-    g_signal_connect (view, "set-scroll-adjustments",
-                      G_CALLBACK (view_scroll_adjustments), NULL);
+    g_signal_connect (view, "notify::hadjustment",
+                      G_CALLBACK (view_set_hadjustment), NULL);
+    g_signal_connect (view, "notify::vadjustment",
+                      G_CALLBACK (view_set_vadjustment), NULL);
     g_signal_connect (view, "populate-popup",
                       G_CALLBACK (view_populate_popup), NULL);
+    g_signal_connect (view, "script-alert",
+                      G_CALLBACK (view_script_alert), NULL);
 
     priv->action_group = gtk_action_group_new ("YelpView");
     gtk_action_group_set_translation_domain (priv->action_group, GETTEXT_PACKAGE);
@@ -282,6 +315,11 @@ yelp_view_dispose (GObject *object)
         priv->document = NULL;
     }
 
+    while (priv->link_actions) {
+        action_entry_free (priv->link_actions->data);
+        priv->link_actions = g_slist_delete_link (priv->link_actions, priv->link_actions);
+    }
+
     priv->back_cur = NULL;
     while (priv->back_list) {
         back_entry_free ((YelpBackEntry *) priv->back_list->data);
@@ -297,7 +335,9 @@ yelp_view_finalize (GObject *object)
     YelpViewPrivate *priv = GET_PRIV (object);
 
     g_free (priv->popup_link_uri);
+    g_free (priv->popup_link_text);
     g_free (priv->popup_image_uri);
+    g_free (priv->popup_code_text);
 
     g_free (priv->page_id);
     g_free (priv->root_title);
@@ -319,6 +359,7 @@ yelp_view_class_init (YelpViewClass *klass)
     nautilus_sendto = g_find_program_in_path ("nautilus-sendto");
 
     websettings = webkit_web_settings_new ();
+    g_object_set (websettings, "enable-universal-access-from-file-uris", TRUE, NULL);
     g_signal_connect (settings,
                       "fonts-changed",
                       G_CALLBACK (settings_set_fonts),
@@ -605,6 +646,46 @@ yelp_view_get_action_group (YelpView *view)
 
 /******************************************************************************/
 
+void
+yelp_view_add_link_action (YelpView                *view,
+                           GtkAction               *action,
+                           YelpViewActionValidFunc  func,
+                           gpointer                 data)
+{
+    YelpActionEntry *entry;
+    YelpViewPrivate *priv = GET_PRIV (view);
+
+    entry = g_new0 (YelpActionEntry, 1);
+    entry->action = g_object_ref (action);
+    entry->func = func;
+    entry->data = data;
+
+    priv->link_actions = g_slist_append (priv->link_actions, entry);
+}
+
+YelpUri *
+yelp_view_get_active_link_uri (YelpView *view)
+{
+    YelpViewPrivate *priv = GET_PRIV (view);
+    YelpUri *uri;
+
+    if (g_str_has_prefix (priv->popup_link_uri, BOGUS_URI))
+        uri = yelp_uri_new_relative (priv->uri, priv->popup_link_uri + BOGUS_URI_LEN);
+    else
+        uri = yelp_uri_new_relative (priv->uri, priv->popup_link_uri);
+
+    return uri;
+}
+
+gchar *
+yelp_view_get_active_link_text (YelpView *view)
+{
+    YelpViewPrivate *priv = GET_PRIV (view);
+    return g_strdup (priv->popup_link_text);
+}
+
+/******************************************************************************/
+
 static void
 view_scrolled (GtkAdjustment *adjustment,
                YelpView      *view)
@@ -619,28 +700,33 @@ view_scrolled (GtkAdjustment *adjustment,
 }
 
 static void
-view_scroll_adjustments (YelpView      *view,
-                         GtkAdjustment *hadj,
-                         GtkAdjustment *vadj,
-                         gpointer       data)
+view_set_hadjustment (YelpView      *view,
+                      GParamSpec    *pspec,
+                      gpointer       data)
 {
     YelpViewPrivate *priv = GET_PRIV (view);
-    priv->vadjustment = vadj;
-    if (priv->vadjuster > 0)
-        g_source_remove (priv->vadjuster);
-    priv->vadjuster = 0;
-    if (vadj) {
-        priv->vadjuster = g_signal_connect (vadj, "value-changed",
-                                            G_CALLBACK (view_scrolled), view);
-    }
-    priv->hadjustment = hadj;
+    priv->hadjustment = gtk_scrollable_get_hadjustment (GTK_SCROLLABLE (view));
     if (priv->hadjuster > 0)
         g_source_remove (priv->hadjuster);
     priv->hadjuster = 0;
-    if (hadj) {
-        priv->hadjuster = g_signal_connect (hadj, "value-changed",
+    if (priv->hadjustment)
+        priv->hadjuster = g_signal_connect (priv->hadjustment, "value-changed",
+                                            G_CALLBACK (view_scrolled), view);
+}
+
+static void
+view_set_vadjustment (YelpView      *view,
+                      GParamSpec    *pspec,
+                      gpointer       data)
+{
+    YelpViewPrivate *priv = GET_PRIV (view);
+    priv->vadjustment = gtk_scrollable_get_vadjustment (GTK_SCROLLABLE (view));
+    if (priv->vadjuster > 0)
+        g_source_remove (priv->vadjuster);
+    priv->vadjuster = 0;
+    if (priv->vadjustment)
+        priv->vadjuster = g_signal_connect (priv->vadjustment, "value-changed",
                                             G_CALLBACK (view_scrolled), view);
-    }
 }
 
 static void
@@ -659,6 +745,9 @@ popup_open_link (GtkMenuItem *item,
 
     g_free (priv->popup_link_uri);
     priv->popup_link_uri = NULL;
+
+    g_free (priv->popup_link_text);
+    priv->popup_link_text = NULL;
 }
 
 static void
@@ -676,6 +765,9 @@ popup_open_link_new (GtkMenuItem *item,
     g_free (priv->popup_link_uri);
     priv->popup_link_uri = NULL;
 
+    g_free (priv->popup_link_text);
+    priv->popup_link_text = NULL;
+
     g_signal_emit (view, signals[NEW_VIEW_REQUESTED], 0, uri);
     g_object_unref (uri);
 }
@@ -695,7 +787,7 @@ file_copied (GFile        *file,
 {
     GError *error = NULL;
     if (!g_file_copy_finish (file, res, &error)) {
-        GtkWidget *dialog = gtk_message_dialog_new (GTK_WIDGET_VISIBLE (data->window) ? data->window : NULL,
+        GtkWidget *dialog = gtk_message_dialog_new (gtk_widget_get_visible (GTK_WIDGET (data->window)) ? data->window : NULL,
                                                     GTK_DIALOG_DESTROY_WITH_PARENT,
                                                     GTK_MESSAGE_ERROR,
                                                     GTK_BUTTONS_OK,
@@ -781,6 +873,97 @@ popup_send_image (GtkMenuItem *item,
     priv->popup_image_uri = NULL;
 }
 
+static void
+popup_copy_code (GtkMenuItem *item,
+                 YelpView    *view)
+{
+    YelpViewPrivate *priv = GET_PRIV (view);
+    GtkClipboard *clipboard = gtk_clipboard_get (GDK_SELECTION_CLIPBOARD);
+    gchar *content = webkit_dom_node_get_text_content (priv->popup_code_node);
+    gtk_clipboard_set_text (clipboard, content, -1);
+    g_free (content);
+}
+
+static void
+popup_save_code (GtkMenuItem *item,
+                 YelpView    *view)
+{
+    YelpViewPrivate *priv = GET_PRIV (view);
+    GtkWidget *dialog, *window;
+    gint res;
+
+    g_free (priv->popup_code_text);
+    priv->popup_code_text = webkit_dom_node_get_text_content (priv->popup_code_node);
+    if (!g_str_has_suffix (priv->popup_code_text, "\n")) {
+        gchar *tmp = g_strconcat (priv->popup_code_text, "\n", NULL);
+        g_free (priv->popup_code_text);
+        priv->popup_code_text = tmp;
+    }
+
+    for (window = gtk_widget_get_parent (GTK_WIDGET (view));
+         window && !GTK_IS_WINDOW (window);
+         window = gtk_widget_get_parent (window));
+
+    dialog = gtk_file_chooser_dialog_new (_("Save Code"),
+                                          GTK_WINDOW (window),
+                                          GTK_FILE_CHOOSER_ACTION_SAVE,
+                                          GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
+                                          GTK_STOCK_SAVE, GTK_RESPONSE_OK,
+                                          NULL);
+    gtk_file_chooser_set_do_overwrite_confirmation (GTK_FILE_CHOOSER (dialog), TRUE);
+    if (priv->popup_code_title) {
+        gchar *filename = webkit_dom_node_get_text_content (priv->popup_code_title);
+        gtk_file_chooser_set_current_name (GTK_FILE_CHOOSER (dialog), filename);
+        g_free (filename);
+    }
+    gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (dialog),
+                                         g_get_user_special_dir (G_USER_DIRECTORY_DOCUMENTS));
+
+    res = gtk_dialog_run (GTK_DIALOG (dialog));
+
+    if (res == GTK_RESPONSE_OK) {
+        GError *error = NULL;
+        GFile *file = gtk_file_chooser_get_file (GTK_FILE_CHOOSER (dialog));
+        GFileOutputStream *stream = g_file_replace (file, NULL, FALSE,
+                                                    G_FILE_CREATE_NONE,
+                                                    NULL,
+                                                    &error);
+        if (stream == NULL) {
+            GtkWidget *dialog = gtk_message_dialog_new (gtk_widget_get_visible (window) ? GTK_WINDOW (window) : NULL,
+                                                        GTK_DIALOG_DESTROY_WITH_PARENT,
+                                                        GTK_MESSAGE_ERROR,
+                                                        GTK_BUTTONS_OK,
+                                                        "%s", error->message);
+            gtk_dialog_run (GTK_DIALOG (dialog));
+            gtk_widget_destroy (dialog);
+            g_error_free (error);
+        }
+        else {
+            /* FIXME: we should do this async */
+            GDataOutputStream *datastream = g_data_output_stream_new (G_OUTPUT_STREAM (stream));
+            if (!g_data_output_stream_put_string (datastream, priv->popup_code_text, NULL, &error)) {
+                GtkWidget *dialog = gtk_message_dialog_new (gtk_widget_get_visible (window) ? GTK_WINDOW (window) : NULL,
+                                                            GTK_DIALOG_DESTROY_WITH_PARENT,
+                                                            GTK_MESSAGE_ERROR,
+                                                            GTK_BUTTONS_OK,
+                                                            "%s", error->message);
+                gtk_dialog_run (GTK_DIALOG (dialog));
+                gtk_widget_destroy (dialog);
+                g_error_free (error);
+            }
+            g_object_unref (datastream);
+        }
+        g_object_unref (file);
+    }
+
+    priv->popup_code_node = NULL;
+    priv->popup_code_title = NULL;
+    g_free (priv->popup_code_text);
+    priv->popup_code_text = NULL;
+
+    gtk_widget_destroy (dialog);
+}
+
 static void
 view_populate_popup (YelpView *view,
                      GtkMenu  *menu,
@@ -792,6 +975,7 @@ view_populate_popup (YelpView *view,
     YelpViewPrivate *priv = GET_PRIV (view);
     GList *children;
     GtkWidget *item;
+    WebKitDOMNode *node, *cur, *link_node = NULL, *code_node = NULL, *code_title_node = NULL;
 
     children = gtk_container_get_children (GTK_CONTAINER (menu));
     while (children) {
@@ -804,7 +988,36 @@ view_populate_popup (YelpView *view,
     event = gtk_get_current_event ();
 
     result = webkit_web_view_get_hit_test_result (WEBKIT_WEB_VIEW (view), (GdkEventButton *) event);
-    g_object_get (result, "context", &context, NULL);
+    g_object_get (result,
+                  "context", &context,
+                  "inner-node", &node,
+                  NULL);
+    for (cur = node; cur != NULL; cur = webkit_dom_node_get_parent_node (cur)) {
+        if (WEBKIT_DOM_IS_ELEMENT (cur) &&
+            webkit_dom_element_webkit_matches_selector ((WebKitDOMElement *) cur,
+                                                        "a", NULL))
+            link_node = cur;
+
+        if (WEBKIT_DOM_IS_ELEMENT (cur) &&
+            webkit_dom_element_webkit_matches_selector ((WebKitDOMElement *) cur,
+                                                        "div.code", NULL)) {
+            WebKitDOMNode *title;
+            code_node = (WebKitDOMNode *)
+                webkit_dom_element_query_selector ((WebKitDOMElement *) cur,
+                                                   "pre.contents", NULL);
+            title = webkit_dom_node_get_parent_node (cur);
+            if (title != NULL && WEBKIT_DOM_IS_ELEMENT (title) &&
+                webkit_dom_element_webkit_matches_selector ((WebKitDOMElement *) title,
+                                                            "div.contents", NULL)) {
+                title = webkit_dom_node_get_previous_sibling (title);
+                if (title != NULL && WEBKIT_DOM_IS_ELEMENT (title) &&
+                    webkit_dom_element_webkit_matches_selector ((WebKitDOMElement *) title,
+                                                                "div.title", NULL)) {
+                    code_title_node = title;
+                }
+            }
+        }
+    }
 
     if (context & WEBKIT_HIT_TEST_RESULT_CONTEXT_LINK) {
         gchar *uri;
@@ -812,6 +1025,47 @@ view_populate_popup (YelpView *view,
         g_free (priv->popup_link_uri);
         priv->popup_link_uri = uri;
 
+        g_free (priv->popup_link_text);
+        priv->popup_link_text = NULL;
+        if (link_node != NULL) {
+            WebKitDOMNode *child;
+            gchar *tmp;
+            gint i, tmpi;
+            gboolean ws;
+
+            child = (WebKitDOMNode *)
+                webkit_dom_element_query_selector (WEBKIT_DOM_ELEMENT (link_node),
+                                                   "div.linkdiv div.title", NULL);
+            if (child != NULL)
+                priv->popup_link_text = webkit_dom_node_get_text_content (child);
+
+            if (priv->popup_link_text == NULL)
+                priv->popup_link_text = webkit_dom_node_get_text_content (link_node);
+
+            tmp = g_new0 (gchar, strlen(priv->popup_link_text) + 1);
+            ws = FALSE;
+            for (i = 0, tmpi = 0; priv->popup_link_text[i] != '\0'; i++) {
+                if (priv->popup_link_text[i] == ' ' || priv->popup_link_text[i] == '\n') {
+                    if (!ws) {
+                        tmp[tmpi] = ' ';
+                        tmpi++;
+                        ws = TRUE;
+                    }
+                }
+                else {
+                    tmp[tmpi] = priv->popup_link_text[i];
+                    tmpi++;
+                    ws = FALSE;
+                }
+            }
+            tmp[tmpi] = '\0';
+            g_free (priv->popup_link_text);
+            priv->popup_link_text = tmp;
+        }
+        else {
+            priv->popup_link_text = g_strdup (uri);
+        }
+
         if (g_str_has_prefix (priv->popup_link_uri, "mailto:")) {
             /* Not using a mnemonic because underscores are common in email
              * addresses, and we'd have to escape them. There doesn't seem
@@ -827,6 +1081,8 @@ view_populate_popup (YelpView *view,
             gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
         }
         else {
+            GSList *cur;
+
             item = gtk_menu_item_new_with_mnemonic (_("_Open Link"));
             g_signal_connect (item, "activate",
                               G_CALLBACK (popup_open_link), view);
@@ -836,6 +1092,21 @@ view_populate_popup (YelpView *view,
             g_signal_connect (item, "activate",
                               G_CALLBACK (popup_open_link_new), view);
             gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
+
+            for (cur = priv->link_actions; cur != NULL; cur = cur->next) {
+                gboolean add;
+                YelpActionEntry *entry = (YelpActionEntry *) cur->data;
+                if (entry->func == NULL)
+                    add = TRUE;
+                else
+                    add = (* entry->func) (view, entry->action,
+                                           priv->popup_link_uri,
+                                           entry->data);
+                if (add) {
+                    item = gtk_action_create_menu_item (entry->action);
+                    gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
+                }
+            }
         }
     }
     else {
@@ -896,11 +1167,38 @@ view_populate_popup (YelpView *view,
         gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
     }
 
+    if (code_node != NULL) {
+        item = gtk_separator_menu_item_new ();
+        gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
+
+        priv->popup_code_node = code_node;
+        priv->popup_code_title = code_title_node;
+
+        item = gtk_menu_item_new_with_mnemonic (_("C_opy Code Block"));
+        g_signal_connect (item, "activate",
+                          G_CALLBACK (popup_copy_code), view);
+        gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
+
+        item = gtk_menu_item_new_with_mnemonic (_("Save Code _Block As..."));
+        g_signal_connect (item, "activate",
+                          G_CALLBACK (popup_save_code), view);
+        gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
+    }
+
     g_object_unref (result);
     gdk_event_free (event);
     gtk_widget_show_all (GTK_WIDGET (menu));
 }
 
+static void
+view_script_alert (YelpView        *view,
+                   WebKitWebFrame  *frame,
+                   gchar           *message,
+                   gpointer         data)
+{
+    printf ("\n\n===ALERT===\n%s\n\n", message);
+}
+
 static gboolean
 view_navigation_requested (WebKitWebView             *view,
                            WebKitWebFrame            *frame,
@@ -938,9 +1236,6 @@ view_resource_request (WebKitWebView         *view,
     gchar last;
     gchar *newpath;
 
-    debug_print (DB_FUNCTION, "entering\n");
-    debug_print (DB_ARG, "    uri=\"%s\"\n", requri);
-
     if (!g_str_has_prefix (requri, BOGUS_URI))
         return;
 
@@ -1233,10 +1528,17 @@ uri_resolved (YelpUri  *uri,
         return;
     case YELP_URI_DOCUMENT_TYPE_NOT_FOUND:
         struri = yelp_uri_get_canonical_uri (uri);
-        error = g_error_new (YELP_ERROR, YELP_ERROR_NOT_FOUND,
-                             _("The URI ‘%s’ does point to a valid page."),
-                             struri);
-        g_free (struri);
+        if (struri != NULL) {
+            error = g_error_new (YELP_ERROR, YELP_ERROR_NOT_FOUND,
+                                 _("The URI ‘%s’ does not point to a valid page."),
+                                 struri);
+            g_free (struri);
+        }
+        else {
+            error = g_error_new (YELP_ERROR, YELP_ERROR_NOT_FOUND,
+                                 _("The URI does not point to a valid page."),
+                                 struri);
+        }
         view_show_error_page (view, error);
         return;
     case YELP_URI_DOCUMENT_TYPE_ERROR:
